VALID_LATITUDE_NAMES = [
    "latitude",
    "lats",
    "lat",
    "Latitude",
    "Lats",
    "Lat",
    "LATITUDE",
    "LATS",
    "LAT",
]

VALID_LONGITUDE_NAMES = ["lon", "lons", "longitude", "longitudes"]
VALID_CARTESIAN_X_NAMES = ["x", "x_coord"]
VALID_CARTESIAN_Y_NAMES = ["y", "y_coord"]

VALID_KERNELS = ["bisquare", "gaussian", "exponential"]
VALID_METRICS = ["euclidean", "haversine"]


MULTIPLE_TESTS = [
    "bonferroni",
    "sidak",
    "holm-sidak",
    "holm",
    "simes-hochberg",
    "hommel",
    "fdr_bh",
    "fdr_by",
    "fdr_tsbh",
    "fdr_tsbky",
]


AVG_EARTH_RADIUS = 6371.0  # in km
