
html[data-theme="light"] {
    --pst-color-primary: #ea7317;
    --pst-color-secondary: #ea7217b2;
    --pst-color-info: #F9D264;
    --pst-color-warning: #F9D264;
}

html[data-theme="dark"] {
    --pst-color-primary: #ea7317;
    --pst-color-secondary: #ea7217b2;
    --pst-color-accent: #ea7217b2;
    --pst-color-info: #F9D264;
    --pst-color-target: #F9D264;
    --pst-color-warning: #F9D264;
}


::-moz-selection { /* Code for Firefox */
    color: #000000;
    background: #ea7317;
}    
::selection {
    color: #000000;
    background: #ea7317;
}


/* INFO admonition */
div.admonition.info {
    border-color: #286279;
}
div.admonition.info > .admonition-title:before {
    background-color: #286279;
}
div.admonition.info > .admonition-title:after {
    color: #286279;
}

/* Class for centering elements */
.center h1 { text-align: center; }
.main-container .navbar-brand {
    justify-content: center;
    flex: 1;
}